;;;;
;;;; ecl-readline.lisp
;;;;
;;;; Copyright (C) 2008 Jason Aquilina <jfa7600@yahoo.com.au>
;;;;
;;;; This program is free software; you can redistribute it and/or
;;;; modify it under the terms of the GNU General Public License
;;;; as published by the Free Software Foundation; either version 2
;;;; of the License, or (at your option) any later version.
;;;;
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;;;;

;;;;
;;;; Readline support for ECL
;;;;
;;;; Ideas and code stolen from:
;;;;    CL-READLINE by Nikodemus Siivola <nikodemus@random-state.net>
;;;;    SB-READLINE by Miles Egan <miles@caddr.com>
;;;;

(defpackage "ECL-READLINE"
  (:use "COMMON-LISP" "UFFI")
  (:export "ENABLE"))

(in-package "ECL-READLINE")

(defvar *history-size* 100)
(defvar *history-file* (concatenate 'string (si::getenv "HOME") "/.ecl-history"))
(defvar *prompt* "")
(defvar *newline* (string #\newline))
(defvar *use-package-nicknames* t)
(defvar *line-number* 1)
(defvar *current-package* nil)

(ffi:clines "#include <stdio.h>")
(ffi:clines "#include <readline/readline.h>")
(load-foreign-library "readline" :system-library t)

(uffi:def-function ("readline" rl-readline) ((prompt :cstring)) :returning :cstring)
(uffi:def-function ("add_history" rl-add-history) ((line :cstring)) :returning :void)
(uffi:def-function ("write_history" rl-write-history) ((line :cstring)) :returning :int)
(uffi:def-function ("stifle_history" rl-stifle-history) ((size :int)) :returning :void)
(uffi:def-function ("read_history" rl-read-history) ((filename :cstring)) :returning :int)

(defclass input-stream (gray::fundamental-character-input-stream)
  ((in-buffer :initform (make-string 0))
   (in-index :initform 0)
   (out-buffer :initform (make-array 0 :element-type 'character :adjustable t :fill-pointer t))))

(defmethod gray::stream-input-p ((stream input-stream))
  t)

(defmethod gray::stream-output-p ((stream input-stream))
  nil)

(defmethod gray::stream-read-char ((stream input-stream))
  (ensure-stream-data stream)
  (with-slots (in-buffer in-index) stream
    (let ((c (char in-buffer in-index)))
      (incf in-index)
      c)))

(defmethod gray::stream-unread-char ((stream input-stream) character)
  (with-slots (in-index) stream
    (if (> in-index 0)
	(decf in-index))))

(defmethod gray::stream-listen ((stream input-stream))
  (with-slots (in-buffer in-index) stream
    (< in-index (length in-buffer))))

(defmethod gray::stream-clear-input ((stream input-stream))
  nil)

(defmethod gray::stream-close ((stream input-stream) &key abort)
  nil)

(defmethod gray::stream-peek-char ((stream input-stream))
  (ensure-stream-data stream)
  (with-slots (in-buffer in-index) stream
    (char in-buffer in-index)))

(defun ensure-stream-data (stream)
  (with-slots (in-buffer in-index) stream
    (if (= in-index (length in-buffer))
	(setq in-buffer (readline)
	      in-index 0))))

(defun current-package-name ()
  (if *use-package-nicknames*
      (if (eq *package* (find-package "COMMON-LISP-USER"))
	  "CL-USER"
	  (car (sort
		(list* (package-name *package*) (package-nicknames *package*))
		(lambda (x y) (< (length x) (length y))))))
      (package-name *package*)))

(defun current-level ()
  (- si::*tpl-level* si::*step-level* -1))

(defun update-symbol-completions ()
  (unless (eq *package* *current-package*)
    (setq *current-package* *package*)
    (let ((entries (make-array 0 :adjustable t :fill-pointer t)))
      (do-symbols (s (find-package *current-package*))
	(vector-push-extend (string-downcase (string s)) entries))
      (sort entries #'string<)
      (c-allocate-completions (length entries))
      (loop for entry across entries
	   do (c-add-completion-entry entry)))))
    
(defun prompt ()
  (fresh-line)
  (update-symbol-completions)
  (setq *prompt*
	(if (= (current-level) 1)
	    (format nil "~A[~A]> " (current-package-name) *line-number*)
	    (format nil "** BREAK [LEVEL ~A] > " (current-level)))))

(defun readline ()
  (let* ((c-line (rl-readline *prompt*))
	 (line (convert-from-cstring c-line)))
    (if (= (current-level) 1)
	(incf *line-number*))
    (setq *prompt* "")
    (rl-add-history c-line)
    (free-cstring c-line)
    (rl-write-history *history-file*)
    (concatenate 'string line *newline*)))

(defun enable (&key ((:history-file filename) nil filename-p) ((:history-size size) nil size-p))
  (when filename-p
    (setq *history-file* filename))
  (when size-p
    (setq *history-size* size))
  (rl-read-history *history-file*)
  (rl-stifle-history *history-size*)
  (c-use-symbol-completer)
  (setf system::*tpl-prompt-hook* #'prompt)
  (setf system::*terminal-io*
	(make-two-way-stream (make-instance 'input-stream) (two-way-stream-output-stream *terminal-io*))))

;;;; End of File